@@ -4,7 +4,7 @@
 #   file 'LICENSE', which is part of this source code package.
 
 
-from typing import Tuple, Dict, List, Optional, Any
+from typing import Tuple, Dict, List, Optional
 from collections import defaultdict, Counter
 from statistics import mean
 import pysam
@@ -13,9 +13,11 @@ import os.path
 import gzip
 import tarfile
 import pickle
+import tempfile
+import indelpost
 
 from .common import log, GRange, AldyException, script_path, Timing, chr_prefix
-from .gene import Gene, Mutation, CNConfigType
+from .gene import Gene, CNConfigType
 from .coverage import Coverage
 from .profile import Profile
 
@@ -26,14 +28,14 @@ class Sample:
     def __init__(
         self,
         gene: Gene,
-        profile: Profile,
+        profile: Optional[Profile],
         path: str,
         reference: Optional[str] = None,
         debug: Optional[str] = None,
     ):
         """
         :param gene: Gene instance.
-        :param profile: Profile instance.
+        :param profile: Profile instance. `None` if loading a dump file.
         :param path: Path to a SAM/BAM/CRAM/VCF/dump file.
         :param reference: Reference genome path for reading CRAM files.
             Default: None.
@@ -62,18 +64,10 @@ class Sample:
         self._dump_reads: List[Tuple[Tuple, List]] = []
         """list[tuple[tuple, list]]: Read information."""
 
-        self._insertion_sites = {
-            m for a in gene.alleles.values() for m in a.func_muts if m.op[:3] == "ins"
-        }  # TODO: currently considers only functional indels
-        """Indel sites whose coverage should be corrected."""
-
-        self._insertion_counts: Dict[Tuple[int, int], int] = defaultdict(int)
-        """Indel coverage (number of reads covering a given indel)."""
-
-        self._insertion_reads: Dict[Mutation, Dict[Any, int]] = {
-            m: defaultdict(int) for m in self._insertion_sites
+        self._indel_sites = {
+            (pos, op): [0, 0] for pos, op in gene.mutations if op[:3] in ["ins", "del"]
         }
-        """Indel coverage (number of reads covering a given indel)."""
+        self._indel_sites_eqs = {}
 
         self._multi_sites = {
             m.pos: m.op
@@ -83,7 +77,7 @@ class Sample:
         }
         """Multi-substitutions (e.g., `A.C>T.G`)."""
 
-        self.grid_columns = {
+        self.phaseable = {
             pos: i for i, pos in enumerate(sorted({pos for pos, _ in gene.mutations}))
         }
         """Locations that should be phased."""
@@ -91,9 +85,6 @@ class Sample:
         self.phases: Dict[str, Dict[int, str]] = {}
         """Phasing information."""
 
-        self.moved: Dict = {}  # TODO: get rid of
-        """Phasing information for corrected/moved indels."""
-
         self._fusion_counter: Dict = {}
         """Fusion read coverage (for long reads)."""
 
@@ -101,8 +92,8 @@ class Sample:
         """Set if long-read data is used."""
 
         with Timing("[sam] Read SAM"):
-            group_indels = False
-            self.kind, self.genome = detect_genome(path)
+            self.kind, _ = detect_genome(path)
+            self.genome = gene.genome
 
             if self.kind == "vcf":
                 try:
@@ -111,22 +102,21 @@ class Sample:
                     raise AldyException(f"VCF {path} is not indexed")
             elif self.kind == "dump":
                 norm, muts = self._load_dump(path)
-                group_indels = True
             else:
-                if self.profile.sam_long_reads:
+                if self.profile and self.profile.sam_long_reads:
                     self.is_long_read = True
                     norm, muts = self._load_long_sam(path, reference, debug)
                 else:
                     norm, muts = self._load_sam(path, reference, debug)
-                    group_indels = True
-                if self.profile.cn_region:
+                if self.profile and self.profile.cn_region:
                     self._dump_cn = self._load_cn_region(
                         path, reference, self.profile.cn_region
                     )
-            self._make_coverage(norm, muts, group_indels=group_indels)
+            self._make_coverage(norm, muts)
             if self.kind == "sam" and debug:
                 self._dump_alignments(f"{debug}.{gene.name}", norm, muts)
 
+        assert self.profile, "profile not set"
         if self.profile.cn_region:
             self.coverage._normalize_coverage()
         log.debug("[sam] avg_coverage= {:.1f}x", self.coverage.average_coverage())
@@ -146,6 +136,7 @@ class Sample:
         log.debug("[sam] path= {}", os.path.abspath(sam_path))
         if reference:
             log.debug("[sam] reference= {}", os.path.abspath(reference))
+        assert self.profile, "profile not set"
 
         norm: dict = defaultdict(list)
         muts: dict = defaultdict(list)
@@ -166,6 +157,8 @@ class Sample:
                 self.gene.chr, [x["SN"] for x in sam.header["SQ"]]
             )
 
+            with tempfile.TemporaryDirectory() as tmp:
+                self._realign_indels(tmp, sam, reference)
             if has_index:
                 iter = sam.fetch(
                     region=self.gene.get_wide_region().samtools(prefix=self._prefix)
@@ -308,22 +301,19 @@ class Sample:
         log.warn("Loading debug dump from {}", dump_path)
         (
             self.name,
+            self.profile,
             self._dump_cn,
             norm,
             muts,
             phases,
             self._fusion_counter,
-            self._insertion_reads,
+            self._indel_sites,
         ) = pickle.load(
-            fd  # type: ignore
-        )
+            fd
+        )  # type: ignore
         self.phases = {f"r{i}": v for i, v in enumerate(phases)}
         norm = {p: [q for q, n in c.items() for _ in range(n)] for p, c in norm.items()}
         muts = {p: [q for q, n in c.items() for _ in range(n)] for p, c in muts.items()}
-        for (start, op), quals in muts.items():
-            if op.startswith("ins"):
-                self._insertion_counts[start, len(op) - 3] += len(quals)
-
         return norm, muts
 
     def _dump_alignments(self, debug: str, norm, muts):
@@ -334,16 +324,97 @@ class Sample:
             pickle.dump(
                 (
                     self.name,
+                    self.profile,
                     self._dump_cn,
                     {p: Counter(q) for p, q in norm.items()},
                     {p: Counter(q) for p, q in muts.items()},
                     [v for v in self.phases.values() if len(v) > 1],
                     self._fusion_counter,
-                    self._insertion_reads,
+                    self._indel_sites,  # TODO: remove
                 ),
                 fd,  # type: ignore
             )
 
+    def _realign_indels(self, tmp, sam, reference, long_reads=False):
+        """Realign reads around database indels via indelpost module."""
+
+        assert self.profile, "profile not loaded"
+
+        rname = f"{self._prefix}{self.gene.chr}"
+        if not reference:
+            reference = f"{tmp}/ref.fa"
+            with open(reference, "w") as f:
+                print(f">{rname}", file=f)
+                print("N" * self.gene._lookup_range[0], end="", file=f)
+                print(self.gene._lookup_seq, end="", file=f)
+                sz = sam.get_reference_length(rname)
+                print("N" * (sz - self.gene._lookup_range[1]), file=f)
+            os.system("samtools faidx " + reference)
+        ref = pysam.FastaFile(reference)  # type: ignore
+
+        prev_indel = None
+        for pos, op in sorted(self._indel_sites, key=lambda x: (x[0], -len(x[1]))):
+            p = pos
+            if op.startswith("del"):
+                if "ins" in op:
+                    pd, pi = op[3:].split("ins")
+                    o1, o2 = pd, pi
+                else:
+                    p -= 1
+                    o = self.gene[p]
+                    o1, o2 = o + op[3:], o
+            else:
+                o = self.gene[p]
+                o1, o2 = o, o + op[3:]
+
+            v = indelpost.Variant(rname, p + 1, o1, o2, ref)  # type: ignore
+
+            if long_reads:
+                # Speed-up: just generate equivalent indels, no need for the realignment
+                for ev in v.generate_equivalents():
+                    np, no = ev.pos - 1, ""
+                    if len(ev.ref) < len(ev.alt) and ev.alt.startswith(ev.ref):
+                        np += len(ev.ref)
+                        no = "ins" + ev.alt[len(ev.ref) :]
+                    elif len(ev.ref) > len(ev.alt) and ev.ref.startswith(ev.alt):
+                        np += len(ev.alt)
+                        no = "del" + ev.ref[len(ev.alt) :]
+                    if no:
+                        self._indel_sites_eqs[np, no] = (pos, op)
+                continue
+
+            valn = indelpost.VariantAlignment(  # type: ignore
+                v,
+                sam,
+                mapping_quality_threshold=self.profile.min_mapq,
+                base_quality_threshold=self.profile.min_quality,
+                # needed to account for indel and database errors
+                exact_match_for_shiftable=True,
+            )
+
+            phased = valn.phase()
+            if len(phased.ref) - len(phased.alt) != len(v.ref) - len(v.alt):
+                continue  # HACK: this indicates a subsumed indel
+
+            self._indel_sites[pos, op] = list(valn.count_alleles())
+            on_target = {r.query_name for r in valn.fetch_reads("target")}
+            if (
+                prev_indel
+                and pos == prev_indel[0]
+                and op.startswith("ins")
+                and prev_indel[1].startswith(op)  # handle indel repeats
+            ):
+                # do not consider previously used reads
+                off, on = self._indel_sites[pos, op]
+                x = len(on_target & prev_indel[2])
+                self._indel_sites[pos, op] = [off + x, on - x]
+            if self._indel_sites[pos, op][1]:
+                log.debug(
+                    "[indel] {}:{} -> {}", pos + 1, op, self._indel_sites[pos, op]
+                )
+                prev_indel = (pos, op, on_target)
+        # assert False
+
     def _load_cn_region(self, path, reference, cn_region):
         """
         Load the copy-number neutral coverage data from a SAM/BAM/CRAM file.
@@ -386,7 +457,7 @@ class Sample:
                             start += size
         return self._dump_cn
 
-    def _make_coverage(self, norm, muts, group_indels=True):
+    def _make_coverage(self, norm, muts):
         """Populate coverage data."""
 
         coverage: Dict[int, Dict[str, List]] = dict()
@@ -395,124 +466,28 @@ class Sample:
                 continue
             coverage.setdefault(pos, {})["_"] = cov
         bounds = min(self.gene.chr_to_ref), max(self.gene.chr_to_ref)
-        for m, cov in muts.items():
-            pos, mut = m
+        for (pos, mut), cov in muts.items():
             if pos not in coverage:
                 coverage[pos] = {}
             if not bounds[0] <= pos <= bounds[1] and mut[:3] != "ins":
                 mut = "_"  # ignore mutations outside of the region of interest
             coverage.setdefault(pos, {}).setdefault(mut, []).extend(cov)
-        if group_indels:
-            self._group_indels(coverage)
         for pos, op in self._multi_sites.items():
             if pos in coverage and op in coverage[pos]:
-                log.debug(
-                    f"[sam] multi-SNP {pos}{op} with {len(coverage[pos][op])} reads"
-                )
-        for mut in self._insertion_sites:
-            if mut.pos in coverage and mut.op in coverage[mut.pos]:
-                total = sum(
-                    len(v) for o, v in coverage[mut.pos].items() if o[:3] != "ins"
-                )
-                if total > 0:
-                    corrected = self._correct_ins_coverage(mut, total)
-                    diff = corrected - len(coverage[mut.pos][mut.op])
-                    coverage[mut.pos][mut.op].extend(
-                        [(mean(mq for mq, _ in coverage[mut.pos][mut.op]), 10)] * diff
-                    )
-
+                log.debug(f"[sam] multi-SNP {pos}{op}: {len(coverage[pos][op])} reads")
+        assert self.profile, "profile not set"
         self.coverage = Coverage(
             self.gene,
             self.profile,
             self,
             {p: {m: v for m, v in coverage[p].items() if len(v) > 0} for p in coverage},
+            self._indel_sites,
             self._dump_cn,
         )
         """Sample coverage data."""
 
         return self.coverage
 
-    def _group_indels(self, coverage):
-        """
-        Group ambiguous indels (e.g., ensure that all `insAA` in `AAA...AAA` region
-        are treated equally).
-        """
-
-        # Group ambiguous deletions
-        indels = defaultdict(set)
-        for m in self.gene.mutations:
-            if "del" in m[1]:
-                indels[m[0]].add(m[1])
-        for pos in coverage:
-            for mut in coverage[pos]:
-                if mut[:3] != "del" or "N" in mut[3:]:
-                    continue
-                potential = [pos]
-                sz = len(mut[3:])
-                deleted = self.gene[pos : pos + sz]
-                for p in range(pos - sz, -1, -sz):
-                    if self.gene[p : p + sz] != deleted:
-                        break
-                    potential.append(p)
-                for p in range(pos + sz, max(coverage), sz):
-                    if self.gene[p : p + sz] != deleted:
-                        break
-                    potential.append(p)
-                potential = [
-                    p
-                    for p in set(potential) & set(indels)
-                    for m in indels[p]
-                    if m == mut
-                ]
-                if len(potential) == 1 and potential[0] != pos:
-                    new_pos = potential[0]
-                    log.trace(
-                        f"[sam] relocate {len(coverage[pos][mut])} "
-                        + f"from {pos+1}{mut} to {new_pos+1}{mut}"
-                    )
-                    coverage[new_pos].setdefault(mut, []).extend(coverage[pos][mut])
-                    coverage[pos][mut] = []
-                    self.moved[pos, mut] = (new_pos, mut)
-
-        # Group ambiguous insertions
-        indels = defaultdict(set)
-        for m in self.gene.mutations:
-            if "ins" in m[1]:
-                indels[m[0]].add(m[1])
-        for pos in coverage:
-            for mut in coverage[pos]:
-                if mut[:3] != "ins":
-                    continue
-                inserted, sz = mut[3:], len(mut[3:])
-                potential = []
-                for p in range(pos, -1, -sz):
-                    potential.append(p)
-                    if self.gene[p : p + sz] != inserted:
-                        break
-                for p in range(pos, max(coverage), sz):
-                    potential.append(p)
-                    if self.gene[p : p + sz] != inserted:
-                        break
-                potential = [
-                    p
-                    for p in set(potential) & set(indels)
-                    for m in indels[p]
-                    if m == mut
-                ]
-                if len(potential) == 1 and potential[0] != pos:
-                    new_pos = potential[0]
-                    log.debug(
-                        f"[sam] relocate {len(coverage[pos][mut])} "
-                        + f"from {pos+1}{mut} to {new_pos+1}{mut}"
-                    )
-                    coverage[new_pos].setdefault(mut, []).extend(coverage[pos][mut])
-                    coverage[pos][mut] = []
-
-                    L = len(mut) - 3
-                    self._insertion_counts[new_pos, L] += self._insertion_counts[pos, L]
-                    self._insertion_counts[pos, L] = 0
-                    self.moved[pos, mut] = (new_pos, mut)
-
     def _parse_read(
         self,
         fragment,
@@ -549,13 +524,13 @@ class Sample:
             """
             if q < 2:
                 return int(q)
-            if 2 <= q < 10:
+            if q < 10:
                 return 6
-            if 10 <= q < 20:
+            if q < 20:
                 return 15
-            if 20 <= q < 29:
+            if q < 29:
                 return 25
-            if 30 <= q < 39:
+            if q < 39:
                 return 35
             return 40
 
@@ -566,21 +541,24 @@ class Sample:
         for op, size in cigar:
             if op == 2:  # Deletion
                 mut = (start, "del" + self.gene[start : start + size])
-                muts[mut].append((bin_quality(mq), bin_quality(prev_q)))
+                for i in range(size):
+                    muts[start + i, "-"].append((bin_quality(mq), bin_quality(prev_q)))
                 dump_arr.append(mut)
-                if start in self.grid_columns:
+                if start in self.phaseable:
                     phase[start] = mut[1]
+                if self._indel_sites_eqs and mut in self._indel_sites_eqs:
+                    self._indel_sites[self._indel_sites_eqs[mut]][1] += 1
                 start += size
             elif op == 1:  # Insertion
                 mut = (start, "ins" + seq[s_start : s_start + size])
                 q = mean(qual[s_start : s_start + size]) if qual else prev_q
                 muts[mut].append((bin_quality(mq), bin_quality(q)))
                 prev_q = q
-                # HACK: just store the length due to seq. errors
-                self._insertion_counts[start, size] += 1
                 dump_arr.append(mut)
-                if start in self.grid_columns:
+                if start in self.phaseable:
                     phase[start] = mut[1]
+                if self._indel_sites_eqs and mut in self._indel_sites_eqs:
+                    self._indel_sites[self._indel_sites_eqs[mut]][1] += 1
                 s_start += size
             elif op == 4:  # Soft-clip
                 s_start += size
@@ -594,11 +572,11 @@ class Sample:
                         mut = (start + i, f"{self.gene[start + i]}>{seq[s_start + i]}")
                         dump_arr.append(mut)
                         muts[mut].append((bin_quality(mq), bin_quality(q)))
-                        if start + i in self.grid_columns:
+                        if start + i in self.phaseable:
                             phase[start + i] = mut[1]
                     else:  # We ignore all mutations outside the RefSeq region
                         norm[start + i].append((bin_quality(mq), bin_quality(q)))
-                        if start + i in self.grid_columns:
+                        if start + i in self.phaseable:
                             phase[start + i] = "_"
                     prev_q = q
                 start += size
@@ -624,52 +602,13 @@ class Sample:
                 muts[pos, op].append(
                     (mean(mq for mq, _ in items), mean(q for _, q in items))
                 )
+
+        for pos, op in self._indel_sites:
+            if ref_start <= pos < start:
+                self._indel_sites[pos, op][0] += 1
         read_pos = (ref_start, start, len(seq))
-        for (pos, _), data in self._insertion_reads.items():
-            if pos >= ref_start - 100 and pos < start + 100:
-                data[read_pos] += 1
         return read_pos, dump_arr
 
-    def _correct_ins_coverage(self, mut: Mutation, total):
-        """
-        Fix low coverage of large tandem insertions.
-
-        Targets the cases where reference looks like
-            `...X...`
-        and the donor genome looks like
-            `...XX...` (i.e., `X` is a tandem insertion).
-
-        Any read that covers only one tandem (e.g. read `X...`) will get perfectly
-        aligned (without trigerring an insertion) to the tail of the tandem insertion.
-        This results in under-estimation of the insertion abundance will (as aligner
-        could not assign `I` CIGAR to the correct insertion).
-
-        This function attempts to correct this bias.
-        """
-
-        MARGIN_RATIO = 0.20
-        INSERT_LEN_RESCALE = 10.0
-        INSERT_LEN_AMPLIFY = 3
-
-        orig_cov = self._insertion_counts[mut.pos, len(mut.op) - 3]
-        new_total = 0.0
-        for (orig_start, orig_end, read_len), cov in self._insertion_reads[mut].items():
-            ins_len = len(mut.op) - 3
-            # HACK: This is horrible way of detecting *40 (min 20% on the sides? max??)
-            min_margin = MARGIN_RATIO * max(1, ins_len / INSERT_LEN_RESCALE) * read_len
-            if orig_end >= mut.pos + max(
-                INSERT_LEN_AMPLIFY * ins_len, min_margin
-            ) and orig_start <= mut.pos - max(
-                (INSERT_LEN_AMPLIFY - 1) * ins_len, min_margin
-            ):
-                new_total += cov
-        new_total += orig_cov
-        new_cov = int(total * (orig_cov / new_total))
-        if new_cov > orig_cov:
-            log.debug(f"[sam] rescale {mut} from {orig_cov} to {new_cov}")
-            return new_cov
-        return orig_cov
-
     def _load_long_sam(self, sam_path: str, reference=None, debug=None):
         """
         Load the read, mutation and coverage data from a long-read SAM/BAM/CRAM file.
@@ -682,6 +621,7 @@ class Sample:
         log.debug("[sam] pacbio_path= {}", os.path.abspath(sam_path))
         assert self.genome, "Genome not provided"
         assert self.genome == "hg38", "Only hg38 supported at this moment"
+        assert self.profile, "profile not set"
 
         norm: dict = defaultdict(list)
         muts: dict = defaultdict(list)
@@ -736,6 +676,12 @@ class Sample:
             self._prefix = chr_prefix(
                 self.gene.chr, [x["SN"] for x in sam.header["SQ"]]
             )
+
+            with tempfile.TemporaryDirectory() as tmp:
+                self._realign_indels(tmp, sam, reference, True)
+                for po, (off, on) in self._indel_sites.items():
+                    self._indel_sites[po] = [off - on, on]
+
             wide = self.gene.get_wide_region()
             end = wide.end
             if self.gene.name == "CYP2D6":
@@ -925,8 +871,9 @@ def detect_genome(sam_path: str) -> Tuple[str, Optional[str]]:
             chrs = {x["SN"]: int(x["LN"]) for x in sam.header["SQ"]}
             is_hg19, is_hg38 = True, True
             for c in "1 10 22".split():
-                is_hg19 &= chrs[_prefix + c] == hg19_lens[c]
-                is_hg38 &= chrs[_prefix + c] == hg38_lens[c]
+                if _prefix + c in chrs:
+                    is_hg19 &= chrs[_prefix + c] == hg19_lens[c]
+                    is_hg38 &= chrs[_prefix + c] == hg38_lens[c]
             if is_hg19:
                 return "sam", "hg19"
             elif is_hg38:
